/* Declarations for startup.c, which is generated in Macaulay2/bin */

typedef struct {
     const char *filename, *contents;
     } cached_file;
extern cached_file startupFile;
extern cached_file testStrings[];
extern int num_testStrings;
