#pragma once

#include "MonomialHashTable.hpp"

namespace newf4 {

class MacaulayMatrix
{
 private:
  MonomialHashTable mMonomials;
};

struct Column
{
};

struct Row
{
};

} // end namespace newf4

// Local Variables:
// indent-tabs-mode: nil
// End: