;;; M2-mode.el --- Major mode for editing Macaulay2 source core -*- lexical-binding: t -*-
;; Version: 1.25.11
;; Keywords: languages Macaulay2
;; URL: https://github.com/Macaulay2/M2-emacs
;; Package-Requires: ((emacs "24.1"))

;;; Commentary:
;; This file used to contain most of M2-mode definitions, which have since moved
;; to M2.el. This file remains for compatibility, or for new features.

(require 'M2)

(provide 'M2-mode)

;;; M2-mode.el ends here
