/*-*- c -*-*/

body {
  color: black; 
  background-color: #d8ffff;
  margin : 1.5em; 
  font-family : Arial,sans-serif;
}

h1   { text-align: center; font-size : 125%; }
h2   { font-size : 110%; }
h3   { font-size : 105%; }
h4   { font-size : 100%; margin-left: 8pt; }
ul   { clear: both; } 

a:link { background-color: transparent; color: #0000ee; }
a:visited { background-color: transparent; color: #52188b; }
a:hover { background-color: transparent; color: #009999; }
a:active { background-color: transparent; color: #ee0000; }

table.buttons td { text-align: center; }

table.matrix {
  background-color: #c0ffff;
  border-style: solid;
  border-top-color: #c0ffff;
  border-bottom-color: #50b0b0;
  border-left-color: #a0ffff;
  border-right-color: #50b0b0;
  border-width: 3px;
  border-spacing: 0px;
  text-align: center;
}

table.matrix td {
  text-align: center;
}

table.examples { 
     margin-left: 2em;
     width: 25%;
     background-color: #c0ffff;
     border-style: solid;
     border-top-color: #c0ffff;
     border-bottom-color: #50b0b0;
     border-left-color: #a0ffff;
     border-right-color: #50b0b0;
     border-width: 3px;
     border-spacing: 0px;
     font-size: 100%;		/* formerly was 75% */
     line-height: 110%;		/* 122% seems to be the default, and 100% leaves almost no space between lines */
}
table.examples td {
     border-style: solid;
     border-top-color: #50b0b0;
     border-bottom-color: #c0ffff;
     border-left-color: #50b0b0;
     border-right-color: #a0ffff;
     padding: 1ex;
     border-width: 1px;
}
table.examples pre {
     margin: 0pt;
     padding: 0pt;
}

div.waystouse {
     font-size: 75%;
     line-height: 110%;
}

div.exports {
     font-size: 80%;
     line-height: 110%;
}

dl.element {
  margin: 0px;
}

div.indent {
    border-left:3px solid;
    padding:10px;
}

span.tt {
  font-family: monospace;
}

kbd {
  display: inline-block;
  font-family: monospace;
  padding: 0.25rem;
  line-height: 0.8em;
  vertical-align: middle;
  background-color: #f6f8fa;
  border: solid 0px #afb8c133;
  border-bottom-color: #afb8c133;
  border-radius: 6px;
  box-shadow: inset 0 -1px 0 #afb8c133;
}

div#buttons {display: flex;}
div#buttons div {flex-grow: 1; align-self: flex-end;}
.right {text-align: right;}

/*

 Local Variables:
 compile-command: "make -C $M2BUILDDIR/Macaulay2/m2 "
 End:

 */
