#include "Basis.hpp"

#include "../mat.hpp"
#include "../freemod.hpp"
#include "../mutablemat.hpp"

#if 0

void append_to_basis(newf4::Basis& B, const Matrix* M)
{
}

auto basis_to_M2_matrix(const newf4::Basis & B, const FreeModule* F) -> Matrix*
{
  return nullptr;
}

auto basis_to_M2_mutable_matrix(const newf4::Basis & B, const FreeModule* F) -> MutableMatrix*
{
  return nullptr;
}

#endif

// Local Variables:
// compile-command: "make -C $M2BUILDDIR/Macaulay2/e "
// indent-tabs-mode: nil
// End:
